# TAR commands to create and restore backups
# create
# tar -cP --same-owner -f /tmp/cs_backup/backup.tar.gz /opt/vpnportal/customer_server/database/ /etc/apache2/sites-available/vpnportal-proxy/

# restore
# sudo tar -xP --same-owner -f /tmp/cs_backup/backup.tar.gz

from shutil     import rmtree
from shutil     import copytree
from shutil     import copy2
from os         import makedirs, path, system, remove
import          subprocess, sys

# Default names of backup directory and file
g_dir_cs_backup    = '/tmp/wavpn_backup'
g_file_backup      = 'backup.tar.gz'

# Default directories to backup
g_dir_cs           = '/opt/vpnportal/customer_server/'
g_dir_web          = '/var/www/vpnportal/www/'
g_dir_apache       = '/etc/apache2/'
g_dir_ds           = '/opt/vpnportal/dispatch_server/'

# CS directories to backup
g_cs_dir_backup = [
  g_dir_cs     + 'certificates', 
  g_dir_cs     + 'cli-config-dir', 
  g_dir_cs     + 'database', 
  g_dir_web    + 'db',
  g_dir_apache + 'sites-available/vpnportal-proxy',
  g_dir_apache + 'ssl',
]

# DS Directories to backup
g_ds_dir_backup = [
  g_dir_ds + 'certificates',
  g_dir_ds + 'license',
  g_dir_ds + 'database'
]

# Default TAR params
g_exe_tar            = 'tar'
g_tar_params_backup  = '-cP --same-owner -f'
g_tar_params_restore = '-xP --same-owner -f'

class BackupManager:
  def __init__(self, backup_path = g_dir_cs_backup, backup_file = g_file_backup, quiet = False, force = False):
    self._backup_path = backup_path
    self._backup_file = backup_file
    self._quiet       = quiet
    self._force       = force

  '''
  This method performs backup of CS configuration.
  '''
  def backup_cs(self):
    self._initialize_backup()
    self._backup_dirs(g_cs_dir_backup)

  '''
  This method performs backup of DS configuration.
  '''
  def backup_ds(self):
    self._initialize_backup()
    self._backup_dirs(g_ds_dir_backup)

  '''
  This method performs backup of DS and CS configuration.
  '''
  def backup_all(self):
    self._initialize_backup()
    self._backup_dirs(g_ds_dir_backup + g_cs_dir_backup)

  '''
  This method restores configuration from backup. It restores both CS and
  DS configuration, whatever is present.
  It deletes backup directory if clean_backup is set to True.
  It deletes content of source directories if clean_dirs is set to True.
  '''
  def restore(self, clean_backup = False, clean_dirs = True):
    if clean_dirs == True:
      self._clean_dirs()
    self._restore_dirs()
    if clean_backup == True:
      self._clean_backup_file()

  def _mkdir(self, newpath):
    if not path.exists(newpath):
      makedirs(newpath)

  def _rmdir(self, path):
    cmd = 'rm -rf ' + path + '/*'
    system(cmd)

  def _get_backup_file(self):
    return self._backup_path + '/' + self._backup_file

  def _initialize_backup(self):
    # Check if output file already exists.
    if self._force == False and path.exists(self._get_backup_file()):
      print('File ' + str(self._get_backup_file()) + ' already exists. Exiting.')
      sys.exit(1)

    self._mkdir(self._backup_path)

  def _backup_dirs(self, backup_dirs):
    self._print('Backup file: ' + str(self._get_backup_file()))
    self._print('Backing up following directories:')
    dirs = ''
    for cs_dir in backup_dirs:
      if path.exists(cs_dir):
        dirs += ' ' + cs_dir 
        self._print(str(cs_dir))
    if dirs:
      cmd = g_exe_tar + ' ' + g_tar_params_backup + ' ' + self._get_backup_file() + ' ' + dirs
      system(cmd)

  '''
  This method removes all source directories.
  Why is this method needed: it is called before restoring configuration from 
  backup to remove all content of directories, not only to overwrite backed up files. 
  If there are files in source location that were not backed up, they have to vanish.
  This method checks if the path exists and if it is contained in backup file.
  '''
  def _clean_dirs(self, backup_dirs = g_cs_dir_backup + g_ds_dir_backup):
    # Get content of backup file to check what files are stored in.
    self._print('Cleaning directories:')
    output = self._run_command_output(['tar', 'tf', self._get_backup_file()])

    for cs_dir in backup_dirs:
      if path.exists(cs_dir) and (cs_dir in output):
        self._print(cs_dir)
        self._rmdir(cs_dir)

  def _restore_dirs(self):
    backup_file = self._get_backup_file()
    self._print('Restoring backup from file: ' + str(backup_file))
    if path.exists(backup_file):
      cmd = g_exe_tar + ' ' + g_tar_params_restore + ' ' + backup_file
      system(cmd)

  def _clean_backup_file(self):
    try:
      remove(self._get_backup_file())
    except OSError:
      pass

  '''
  This method runs command and returns it's output.
  '''
  def _run_command_output(self, cmd):
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=False)
    (out, err) = proc.communicate()
    return out.decode()

  def _print(self, message):
    if not self._quiet:
      print(message)

